from pathlib import Path

import pytest
from geopandas import GeoDataFrame  # type: ignore
from shapely.geometry import Point  # type: ignore

from latviastreetify.resolvers import Language, SteetsAndNeighborhoodsResolver


@pytest.fixture
def streets_and_neighborhoods_resolver():
    return SteetsAndNeighborhoodsResolver()


def test_read(streets_and_neighborhoods_resolver):
    result = streets_and_neighborhoods_resolver.read()
    assert isinstance(result, GeoDataFrame)
    assert "geometry" in result.columns
    assert "iela" in result.columns
    assert "adrese" in result.columns
    assert "Code" in result.columns
    assert "Name" in result.columns


def test_get_translation_dict(streets_and_neighborhoods_resolver):
    result = set(streets_and_neighborhoods_resolver.get_translation_dict().values())
    expected_set = set(
        {
            "Nr": "Number",
            "X": "X",
            "Y": "Y",
            "iela": "Street",
            "adrese": "Address",
            "geometry": "geometry",
            "Code": "Code",
            "Name": "Name",
            "lat": "lat",
            "lon": "lon",
        }.values()
    )
    assert result == expected_set


def test_get_gdf(streets_and_neighborhoods_resolver):
    gdf = streets_and_neighborhoods_resolver.get_gdf(language=Language.EN)
    assert isinstance(gdf, GeoDataFrame)
    assert "geometry" in gdf.columns
    assert "Street" in gdf.columns
    assert "Address" in gdf.columns
    assert "Code" in gdf.columns
    assert "Name" in gdf.columns

    gdf_lav = streets_and_neighborhoods_resolver.get_gdf()
    assert isinstance(gdf_lav, GeoDataFrame)
    assert "geometry" in gdf_lav.columns
    assert "iela" in gdf_lav.columns
    assert "adrese" in gdf_lav.columns
    assert "Code" in gdf_lav.columns
    assert "Name" in gdf_lav.columns



"""
module that resolve the streets and neighborhoods of Latvia
"""
from abc import ABC, abstractmethod
from enum import Enum
from os import environ
from pathlib import Path
from typing import Dict

import geopandas as gpd  # type: ignore
import pandas as pd  # type: ignore
import pyproj  # type: ignore


class Language(str, Enum):
    EN = "EN"
    LAV = "LAV"


class Resolver(ABC):
    def __init__(self) -> None:
        self._gdf = self.read()
        super().__init__()

    @abstractmethod
    def read(self) -> gpd.GeoDataFrame:
        ...

    @abstractmethod
    def get_translation_dict(self) -> Dict[str, str]:
        ...

    def get_gdf(self, language: Language = Language.LAV):
        match language:
            case language.EN:
                return self._gdf.rename(columns=self.get_translation_dict())
            case Language.LAV | _:
                return self._gdf


class SteetsResolver(Resolver):
    def read(self) -> gpd.GeoDataFrame:
        if filepath := environ.get("STREETS_FILE"):  # type: ignore
            filepath = Path(filepath)  # type: ignore
            print(f"will use external file {filepath}")
        else:
            filepath = Path(__file__).parent / "data/adreses.shp"  # type: ignore
        gdf = gpd.read_file(filepath).to_crs(3059)
        gdf = gdf.apply(
            lambda col: col.map(lambda s: s.lower() if isinstance(s, str) else s)
        )
        gdf["adrese"] = gdf["adrese"].str.replace(" k-", "k-")
        lks92 = pyproj.CRS.from_epsg(3059)
        wgs84 = pyproj.CRS.from_epsg(4326)
        transformer = pyproj.Transformer.from_crs(lks92, wgs84, always_xy=True)

        def convert_coordinates(row):
            x, y = row["X"], row["Y"]
            lon, lat = transformer.transform(x, y)
            return pd.Series({"lat": lat, "lon": lon})

        gdf[["lat", "lon"]] = gdf.apply(convert_coordinates, axis=1)
        return gdf.dropna()

    def get_translation_dict(self) -> Dict[str, str]:
        translation_dict = {
            "Nr": "Number",
            "X": "X",
            "Y": "Y",
            "lat": "lat",
            "lon": "lon",
            "iela": "Street",
            "adrese": "Address",
            "geometry": "Geometry",
        }
        return translation_dict


class NeighborhoodsResolver(Resolver):
    def read(self) -> gpd.GeoDataFrame:
        if filepath := environ.get("NEIGHBORHOODS_FILE"):  # type: ignore
            filepath = Path(filepath)  # type: ignore
            print(f"will use external file {filepath}")
        else:
            filepath = Path(__file__).parent / "data/Apkaimes.shp"  # type: ignore
        gdf = gpd.read_file(filepath)
        gdf = gdf.apply(
            lambda col: col.map(lambda s: s.lower() if isinstance(s, str) else s)
        )
        return gdf.dropna()

    def get_translation_dict(self) -> Dict[str, str]:
        translation_dict = {
            "Code": "Code",
            "Name": "Name",
            "geometry": "geometry",
        }
        return translation_dict


class SteetsAndNeighborhoodsResolver(Resolver):
    def __init__(self) -> None:
        self._streets_resolver = SteetsResolver()
        self._neighborhoods_resolver = NeighborhoodsResolver()
        super().__init__()

    def read(self) -> gpd.GeoDataFrame:
        streets_gdf = self._streets_resolver.get_gdf()
        neighborhoods_gdf = self._neighborhoods_resolver.get_gdf()
        return streets_gdf.sjoin(
            neighborhoods_gdf, how="left", predicate="within"
        ).dropna()

    def get_translation_dict(self) -> Dict[str, str]:
        all_dict: Dict[str, str] = dict()
        all_dict.update(**self._streets_resolver.get_translation_dict())
        all_dict.update(**self._neighborhoods_resolver.get_translation_dict())
        return all_dict



"""
examples to get the head of the dataframe
"""
from pprint import pprint

from latviastreetify.resolvers import Language, SteetsAndNeighborhoodsResolver

resolver = SteetsAndNeighborhoodsResolver()
gdf = resolver.get_gdf(language=Language.EN)
pprint(gdf.head())



""" module for CLI on latviastreetify
"""

from pprint import pprint

import typer

from latviastreetify.resolvers import Language, SteetsAndNeighborhoodsResolver

app = typer.Typer()


def generate_command(search_column_en, search_column_lav):
    def command(search_value: str, language: Language = Language.EN):
        search_value = search_value.lower()
        sr = SteetsAndNeighborhoodsResolver()
        gdf = sr.get_gdf(language=language)
        match language:
            case Language.EN:
                pprint(gdf[gdf[search_column_en].str.contains(search_value)])
            case Language.LAV | _:
                pprint(gdf[gdf[search_column_lav].str.contains(search_value)])

    return command


SEARCH_COLUMNS = {
    "address": ("Address", "adrese"),
    "name": ("Name", "Name"),
    "street": ("Street", "iela"),
    "number": ("Number", "Nr"),
    "code": ("Code", "Code"),
}

for search_name, (search_column_en, search_column_lav) in SEARCH_COLUMNS.items():
    command = generate_command(search_column_en, search_column_lav)
    app.command(name=f"search-by-{search_name}")(command)


@app.command()
def search_multiple(
    address: str = typer.Option(None),
    name: str = typer.Option(None),
    street: str = typer.Option(None),
    number: str = typer.Option(None),
    code: str = typer.Option(None),
    language: Language = Language.EN,
):
    search_values = {
        "address": address,
        "name": name,
        "street": street,
        "number": number,
        "code": code,
    }

    sr = SteetsAndNeighborhoodsResolver()
    gdf = sr.get_gdf(language=language)
    for search_name, search_value in search_values.items():
        if search_value:
            search_value = search_value.lower()
            match language:
                case Language.EN:
                    gdf = gdf[
                        gdf[SEARCH_COLUMNS[search_name][0]].str.contains(search_value)
                    ]
                case Language.LAV | _:
                    gdf = gdf[
                        gdf[SEARCH_COLUMNS[search_name][1]].str.contains(search_value)
                    ]
    pprint(gdf)


if __name__ == "__main__":
    app()
